//
//  GLMovieViewController.m
//  GLMovie
//
//  Created by Yoshikazu Kuramochi on 11/10/04.
//  Copyright 2011 __MyCompanyName__. All rights reserved.
//

#import <QuartzCore/QuartzCore.h>

#import "GLMovieViewController.h"
#import "EAGLView.h"

@interface GLMovieViewController ()
@property (nonatomic, retain) EAGLContext *context;
@property (nonatomic, assign) CADisplayLink *displayLink;
@end

@implementation GLMovieViewController

@synthesize animating, context, displayLink;


static const NSString* itemStatusContext;

static const GLfloat vertices[] = {
	-0.5,  0.5, -0.5,	 0.5,  0.5, -0.5,	-0.5,  0.5,  0.5,
	 0.5,  0.5,  0.5,	-0.5,  0.5,  0.5,	 0.5,  0.5, -0.5,
	 0.5,  0.5,  0.5,	 0.5,  0.5, -0.5,	 0.5, -0.5,  0.5,
	 0.5, -0.5, -0.5,	 0.5, -0.5,  0.5,	 0.5,  0.5, -0.5,
	 0.5,  0.5, -0.5,	-0.5,  0.5, -0.5,	 0.5, -0.5, -0.5,
	-0.5, -0.5, -0.5,	 0.5, -0.5, -0.5,	-0.5,  0.5, -0.5,
	-0.5,  0.5, -0.5,	-0.5,  0.5,  0.5,	-0.5, -0.5, -0.5,
	-0.5, -0.5,  0.5,	-0.5, -0.5, -0.5,	-0.5,  0.5,  0.5,
	-0.5, -0.5,  0.5,	-0.5,  0.5,  0.5,	 0.5, -0.5,  0.5,
	 0.5,  0.5,  0.5,	 0.5, -0.5,  0.5,	-0.5,  0.5,  0.5,
	-0.5, -0.5,  0.5,	 0.5, -0.5,  0.5,	-0.5, -0.5, -0.5,
	 0.5, -0.5, -0.5,	-0.5, -0.5, -0.5,	 0.5, -0.5,  0.5,
};

static const GLfloat normals[] = {
	 0, 1, 0,	 0, 1, 0,	 0, 1, 0,
	 0, 1, 0,	 0, 1, 0,	 0, 1, 0,
	 1, 0, 0,	 1, 0, 0,	 1, 0, 0,
	 1, 0, 0,	 1, 0, 0,	 1, 0, 0,
	 0, 0,-1,	 0, 0,-1,	 0, 0,-1,
	 0, 0,-1,	 0, 0,-1,	 0, 0,-1,
	-1, 0, 0,	-1, 0, 0,	-1, 0, 0,
	-1, 0, 0,	-1, 0, 0,	-1, 0, 0,
	 0, 0, 1,	 0, 0, 1,	 0, 0, 1,
	 0, 0, 1,	 0, 0, 1,	 0, 0, 1,
	 0,-1, 0,	 0,-1, 0,	 0,-1, 0,
	 0,-1, 0,	 0,-1, 0,	 0,-1, 0,
};

static const GLfloat texCoords[] = {
	0, 0,	1, 0,	0, 1,
	1, 1,	0, 1,	1, 0,
	0, 0,	1, 0,	0, 1,
	1, 1,	0, 1,	1, 0,
	0, 0,	1, 0,	0, 1,
	1, 1,	0, 1,	1, 0,
	0, 0,	1, 0,	0, 1,
	1, 1,	0, 1,	1, 0,
	0, 0,	1, 0,	0, 1,
	1, 1,	0, 1,	1, 0,
	0, 0,	1, 0,	0, 1,
	1, 1,	0, 1,	1, 0,
};


- (void)awakeFromNib
{
    EAGLContext *aContext = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
    if (!aContext)
        NSLog(@"Failed to create ES context");
    else if (![EAGLContext setCurrentContext:aContext])
        NSLog(@"Failed to set ES context current");
    
	self.context = aContext;
	[aContext release];
	
    [(EAGLView *)self.view setContext:context];
    [(EAGLView *)self.view setFramebuffer];
    
    animating = FALSE;
    animationFrameInterval = 1;
    self.displayLink = nil;
	
	
	NSURL* assetURL = [[NSBundle mainBundle] URLForResource:@"iori.mp4" withExtension:nil];
	NSDictionary* assetOptions = [NSDictionary dictionaryWithObject:[NSNumber numberWithBool:YES]
															 forKey:AVURLAssetPreferPreciseDurationAndTimingKey];
	asset = [[AVURLAsset alloc] initWithURL:assetURL options:assetOptions];
	[asset loadValuesAsynchronouslyForKeys:[NSArray arrayWithObject:@"tracks"] completionHandler:^{
		NSError* error = nil;
		AVKeyValueStatus status = [asset statusOfValueForKey:@"tracks" error:&error];
		if (status == AVKeyValueStatusLoaded) {
			playerItem = [[AVPlayerItem alloc] initWithAsset:asset];
			[playerItem addObserver:self forKeyPath:@"status" options:0 context:&itemStatusContext];
			[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(playerItemDidReachEnd:)
														 name:AVPlayerItemDidPlayToEndTimeNotification
													   object:playerItem];
			player = [[AVPlayer alloc] initWithPlayerItem:playerItem];
		} else {
			NSLog(@"%@", [error localizedDescription]);
		}
	}];

	AVAssetTrack* srcTrack = [[asset tracksWithMediaType:AVMediaTypeVideo] objectAtIndex:0];
	comp = [AVMutableComposition composition];
	track = [comp addMutableTrackWithMediaType:AVMediaTypeVideo preferredTrackID:kCMPersistentTrackID_Invalid];
	NSError* error = nil;
	BOOL ok = [track insertTimeRange:srcTrack.timeRange ofTrack:srcTrack atTime:kCMTimeZero error:&error];
	if (!ok) {
		NSLog(@"%@", [error localizedDescription]);
	}
	imageGen = [[AVAssetImageGenerator alloc] initWithAsset:comp];

	buffer = malloc(256 * 256 * 4);
	if (buffer == nil) {
		NSLog(@"out of memory");
	}

	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glEnable(GL_TEXTURE_2D);


	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glRotatef(30, 1, 0, 0);
	glRotatef(30, 0, 1, 0);
	glGetFloatv(GL_MODELVIEW_MATRIX, rotateMatrix);
	glLoadIdentity();
	
	GLfloat light_pos[]      = { 1, 1, 1, 0 };
	GLfloat light_ambient[]  = { 1, 1, 1, 1 };
	glShadeModel(GL_SMOOTH);
	glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
	glClearDepthf(1);
	glEnable(GL_DEPTH_TEST);

	glVertexPointer(3, GL_FLOAT, 0, vertices);
	glEnableClientState(GL_VERTEX_ARRAY);
	glNormalPointer(GL_FLOAT, 0, normals);
	glEnableClientState(GL_NORMAL_ARRAY);
	glTexCoordPointer(2, GL_FLOAT, 0, texCoords);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	zoom = 2;
	panVelocity.x = 0;
	panVelocity.y = 0;

	
	UITapGestureRecognizer* tap2 = [[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(handleTap2:)];
	tap2.numberOfTapsRequired = 2;
	[self.view addGestureRecognizer:tap2];
	[tap2 release];
	
	UITapGestureRecognizer* tap1 = [[UITapGestureRecognizer alloc] initWithTarget:self action:@selector(handleTap1:)];
	tap1.numberOfTapsRequired = 1;
	[tap1 requireGestureRecognizerToFail:tap2];
	[self.view addGestureRecognizer:tap1];
	[tap1 release];

	UIPinchGestureRecognizer* pinch =[[UIPinchGestureRecognizer alloc] initWithTarget:self action:@selector(handlePinch:)];
	[self.view addGestureRecognizer:pinch];
	[pinch release];

	UIPanGestureRecognizer* pan = [[UIPanGestureRecognizer alloc] initWithTarget:self action:@selector(handlePan:)];
	[self.view addGestureRecognizer:pan];
	[pan release];
}

- (void)dealloc
{
    // Tear down context.
    if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];
    
    [context release];
    
	glDeleteTextures(1, &texture);
	CGContextRelease(cgContext);
	free(buffer);
	[imageGen release];
	[track release];
	[comp release];
	[player release];
	[playerItem release];
	[asset release];
	
    [super dealloc];
}

- (void)viewWillAppear:(BOOL)animated
{
    [self startAnimation];
    
    [super viewWillAppear:animated];
}

- (void)viewWillDisappear:(BOOL)animated
{
    [self stopAnimation];
    
    [super viewWillDisappear:animated];
}

- (void)viewDidUnload
{
	[super viewDidUnload];
	
    // Tear down context.
    if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];
	self.context = nil;	
}

- (NSInteger)animationFrameInterval
{
    return animationFrameInterval;
}

- (void)setAnimationFrameInterval:(NSInteger)frameInterval
{
    /*
	 Frame interval defines how many display frames must pass between each time the display link fires.
	 The display link will only fire 30 times a second when the frame internal is two on a display that refreshes 60 times a second. The default frame interval setting of one will fire 60 times a second when the display refreshes at 60 times a second. A frame interval setting of less than one results in undefined behavior.
	 */
    if (frameInterval >= 1)
    {
        animationFrameInterval = frameInterval;
        
        if (animating)
        {
            [self stopAnimation];
            [self startAnimation];
        }
    }
}

- (void)startAnimation
{
    if (!animating)
    {
        CADisplayLink *aDisplayLink = [[UIScreen mainScreen] displayLinkWithTarget:self selector:@selector(drawFrame)];
        [aDisplayLink setFrameInterval:animationFrameInterval];
        [aDisplayLink addToRunLoop:[NSRunLoop currentRunLoop] forMode:NSDefaultRunLoopMode];
        self.displayLink = aDisplayLink;
        
        animating = TRUE;
    }
}

- (void)stopAnimation
{
    if (animating)
    {
        [self.displayLink invalidate];
        self.displayLink = nil;
        animating = FALSE;
    }
}

- (void)drawFrame
{
    [(EAGLView *)self.view setFramebuffer];


	CMTime time = [player currentTime];
	CMTime actualTime;
	NSError* error = nil;
	CGImageRef cgImage = [imageGen copyCGImageAtTime:time actualTime:&actualTime error:&error];
	if (cgImage != nil) {
		size_t width = CGImageGetWidth(cgImage);
		size_t height = CGImageGetHeight(cgImage);
		size_t size = MIN(width, height);
		if (!cgContext) {
			CGBitmapInfo bitmapInfo = kCGBitmapByteOrder32Big;
			switch (CGImageGetAlphaInfo(cgImage)) {
				case kCGImageAlphaPremultipliedLast:
				case kCGImageAlphaPremultipliedFirst:
				case kCGImageAlphaLast:
				case kCGImageAlphaFirst:
					bitmapInfo |= kCGImageAlphaPremultipliedLast;
					break;
				default:
					bitmapInfo |= kCGImageAlphaNoneSkipLast;
					break;
			}
			CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
			cgContext = CGBitmapContextCreate(buffer, 256, 256, 8, 256*4, colorSpace, bitmapInfo);
			CGColorSpaceRelease(colorSpace);
			if (width != 256 || height != 256) {
				CGContextScaleCTM(cgContext, 256.0f/size, 256.0f/size);
			}
		}
		CGRect rect = CGRectMake((width-size)*0.5f, (height-size)*0.5f, size, size);
		CGContextDrawImage(cgContext, rect, cgImage);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 256, 256, 0, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
		//glGenerateMipmap(GL_TEXTURE_2D);
		CGImageRelease(cgImage);
	} else {
		NSLog(@"%@", [error localizedDescription]);
	}

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	double plus = 1/zoom;
	double minus = -plus;
	glFrustumf(minus, plus, minus*1.5, plus*1.5, 5, 15.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	double angle = sqrt(panVelocity.x*panVelocity.x + panVelocity.y*panVelocity.y) / 30;
	glRotatef(angle, panVelocity.y, panVelocity.x, 0);
	glMultMatrixf(rotateMatrix);
	glGetFloatv(GL_MODELVIEW_MATRIX, rotateMatrix);
	glLoadIdentity();
	glTranslatef(0, 0, -10);
	glMultMatrixf(rotateMatrix);
	panVelocity.x = 0;
	panVelocity.y = 0;
	
    glDrawArrays(GL_TRIANGLES, 0, sizeof(vertices)/sizeof(GLfloat)/3);
    
    [(EAGLView *)self.view presentFramebuffer];
}

- (void)didReceiveMemoryWarning
{
    // Releases the view if it doesn't have a superview.
    [super didReceiveMemoryWarning];
    
    // Release any cached data, images, etc. that aren't in use.
}

- (void)observeValueForKeyPath:(NSString*)keyPath ofObject:(id)object 
						change:(NSDictionary*)change context:(void*)ctx
{
	if (ctx == &itemStatusContext) {
		if (playerItem.status == AVPlayerItemStatusReadyToPlay && player.rate == 0) {
			for (AVPlayerItemTrack* itemTrack in playerItem.tracks) {
				if (![itemTrack.assetTrack hasMediaCharacteristic:AVMediaCharacteristicAudible]) {
					itemTrack.enabled = NO;
				}
			}
			[player play];
		}
	} else {
		[super observeValueForKeyPath:keyPath ofObject:object change:change context:ctx];
	}
}

- (void)playerItemDidReachEnd:(NSNotification*)notification
{
	[player seekToTime:kCMTimeZero];
	//[player pause];
}

- (void)handleTap2:(UIGestureRecognizer*)sender
{
	[player pause];
	[player seekToTime:kCMTimeZero];
}

- (void)handleTap1:(UIGestureRecognizer*)sender
{
	if (player.rate == 0) {
		[player play];
	} else {
		CGPoint point = [sender locationInView:self.view];
		CGFloat height = self.view.bounds.size.height;
		if (height*0.1 > point.y) {
			player.rate *= 2;
		} else if (height*0.9 < point.y) {
			player.rate *= 0.5;
		} else {
			[player pause];
		}
	}
}

- (void)handlePinch:(UIGestureRecognizer*)sender
{
	CGFloat scale = ((UIPinchGestureRecognizer*)sender).scale;
	switch (sender.state) {
		case UIGestureRecognizerStateBegan:
			zoom *= scale;
			break;
		case UIGestureRecognizerStateChanged:
			zoom *= scale / prevPinchScale;
			break;
		case UIGestureRecognizerStateEnded:
			zoom *= scale / prevPinchScale;
			zoom = MIN(MAX(zoom, 0.1), 10);
			break;
	}
	prevPinchScale = scale;
}

- (void)handlePan:(UIGestureRecognizer*)sender
{
	CGPoint velocity = [(UIPanGestureRecognizer*)sender velocityInView:self.view];
	switch (sender.state) {
		case UIGestureRecognizerStateBegan:
		case UIGestureRecognizerStateChanged:
		case UIGestureRecognizerStateEnded:
			panVelocity = velocity;
			break;
		default:
			panVelocity.x = 0;
			panVelocity.y = 0;
			break;
	}
}

@end
